#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.09 07:00:00                  #
# ================================================== #

from PySide6.QtGui import Qt
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QCheckBox, QSizePolicy

from pygpt_net.ui.widget.element.labels import HelpLabel, TitleLabel
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.utils import trans


class Prompt:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup system prompt

        :return: QWidget
        """
        # cmd enable/disable
        self.window.ui.nodes['cmd.enabled'] = QCheckBox(trans('cmd.enabled'))
        self.window.ui.nodes['cmd.enabled'].stateChanged.connect(
            lambda: self.window.controller.chat.common.toggle_cmd(self.window.ui.nodes['cmd.enabled'].isChecked())
        )
        self.window.ui.nodes['cmd.enabled'].setToolTip(trans('cmd.tip'))

        # label
        self.window.ui.nodes['toolbox.prompt.label'] = TitleLabel(trans("toolbox.prompt"))

        # clear
        self.window.ui.nodes['preset.clear'] = QPushButton(trans('preset.clear'))
        self.window.ui.nodes['preset.clear'].clicked.connect(
            lambda: self.window.controller.presets.clear())

        # use
        self.window.ui.nodes['preset.use'] = QPushButton(trans('preset.use'))
        self.window.ui.nodes['preset.use'].clicked.connect(
            lambda: self.window.controller.presets.use())
        self.window.ui.nodes['preset.use'].setVisible(False)

        # cols
        header = QHBoxLayout()
        header.addWidget(self.window.ui.nodes['toolbox.prompt.label'])
        header.addWidget(self.window.ui.nodes['cmd.enabled'])
        header.addStretch(1)
        header.addWidget(self.window.ui.nodes['preset.use'], alignment=Qt.AlignRight)
        header.addWidget(self.window.ui.nodes['preset.clear'], alignment=Qt.AlignRight)

        # prompt
        option = self.window.controller.presets.editor.get_option('prompt')
        self.window.ui.nodes['preset.prompt'] = OptionTextarea(self.window, 'preset', 'prompt', option)
        self.window.ui.nodes['preset.prompt'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.window.ui.nodes['preset.prompt'].textChanged.connect(
            lambda: self.on_prompt_changed())

        self.window.ui.nodes['tip.toolbox.prompt'] = HelpLabel(trans('tip.toolbox.prompt'), self.window)

        # rows
        layout = QVBoxLayout()
        layout.addLayout(header)
        layout.addWidget(self.window.ui.nodes['preset.prompt'])
        layout.addWidget(self.window.ui.nodes['tip.toolbox.prompt'])
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        return widget

    def on_prompt_changed(self):
        """On prompt changed"""
        # hook update
        self.window.controller.config.input.on_update(
            self.window.ui.nodes['preset.prompt'].parent_id,
            self.window.ui.nodes['preset.prompt'].id,
            self.window.ui.nodes['preset.prompt'].option,
            self.window.ui.nodes['preset.prompt'].toPlainText(),
        )
        self.window.controller.ui.update_tokens()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.01 02:00:00                  #
# ================================================== #

import os
from typing import Any

from llama_index.readers.google.calendar.base import GoogleCalendarReader as Reader

SCOPES = ["https://www.googleapis.com/auth/calendar.readonly"]

class GoogleCalendarReader(Reader):
    """
    Wrapper for:

    llama_index.readers.google.calendar.base.GoogleCalendarReader

    Extending base reader with credentials configuration.
    """

    def __init__(
        self,
        credentials_path: str = "credentials.json",
        token_path: str = "token.json",
    ) -> None:
        """Initialize with parameters."""
        self.credentials_path = credentials_path
        self.token_path = token_path

    def _get_credentials(self) -> Any:
        """Get valid user credentials from storage.

        ** Based on base reader, extended with credentials configuration. **
        https://github.com/run-llama/llama_index/blob/main/llama-index-integrations/readers/llama-index-readers-google/llama_index/readers/google/calendar/base.py

        The file token.json stores the user's access and refresh tokens, and is
        created automatically when the authorization flow completes for the first
        time.

        Returns:
            Credentials, the obtained credential.
        """
        from google_auth_oauthlib.flow import InstalledAppFlow

        from google.auth.transport.requests import Request
        from google.oauth2.credentials import Credentials

        creds = None
        if os.path.exists(self.token_path):
            creds = Credentials.from_authorized_user_file(self.token_path, SCOPES)

        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.credentials_path, SCOPES
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.token_path, "w") as token:
                token.write(creds.to_json())

        return creds


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.30 20:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class Files:
    def __init__(self, window=None):
        """
        Chat files/attachments controller

        :param window: Window instance
        """
        self.window = window

    def upload(self, mode: str) -> dict:
        """
        Upload attachments

        :param mode: mode
        :return: uploaded attachments list
        """
        self.window.core.gpt.assistants.file_ids = []  # clear file ids
        attachments_list = {}

        if mode == 'assistant':
            is_upload = False
            num_uploaded = 0
            try:
                # upload only new attachments (not uploaded yet to remote)
                attachments = self.window.core.attachments.get_all(mode)
                c = self.window.controller.assistant.files.count_upload(attachments)
                if c > 0:
                    is_upload = True
                    self.window.ui.status(trans('status.uploading'))
                    num_uploaded = self.window.controller.assistant.files.upload(
                        mode,
                        attachments,
                    )
                    self.window.core.gpt.assistants.file_ids = self.window.core.attachments.get_ids(mode)
                    attachments_list = self.window.core.attachments.make_json_list(attachments)

                # show uploaded status
                if is_upload and num_uploaded > 0:
                    self.window.ui.status(trans('status.uploaded'))

            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(str(e))

        return attachments_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.01 02:00:00                  #
# ================================================== #

from llama_index.core.readers.base import BaseReader
from llama_index.readers.web.rss.base import RssReader

from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "rss"
        self.name = "RSS"
        self.type = ["web"]
        self.instructions = [
            {
                "rss": "use it to read RSS feed URL, allowed args: `url`: str",
            }
        ]

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        return RssReader()

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for reader

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        args = {}
        args["urls"] = [kwargs.get("url")]  # list of links
        return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.30 15:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget, QCheckBox

from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.utils import trans


class Agent:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup agent options

        :return: QWidget
        """
        # img variants
        option = self.window.controller.agent.options["agent.iterations"]
        self.window.ui.nodes['agent.iterations.label'] = QLabel(trans("toolbox.agent.iterations.label"))
        self.window.ui.config['global']['agent.iterations'] = \
            OptionSlider(
                self.window,
                'global',
                'agent.iterations',
                option,
            )

        # img raw
        self.window.ui.config['global']['agent.auto_stop'] = QCheckBox(trans("toolbox.agent.auto_stop.label"))
        self.window.ui.config['global']['agent.auto_stop'].stateChanged.connect(
            lambda:
            self.window.controller.agent.toggle_auto_stop(
                self.window.ui.config['global']['agent.auto_stop'].isChecked())
        )

        # label
        label = QLabel(trans("toolbox.agent.iterations.label"))

        # options
        cols = QHBoxLayout()
        cols.addWidget(self.window.ui.config['global']['agent.auto_stop'])
        cols.addWidget(self.window.ui.config['global']['agent.iterations'])

        # rows
        rows = QVBoxLayout()
        rows.addWidget(label)
        rows.addLayout(cols)

        self.window.ui.nodes['agent.options'] = QWidget()
        self.window.ui.nodes['agent.options'].setLayout(rows)
        self.window.ui.nodes['agent.options'].setContentsMargins(0, 0, 0, 0)

        return self.window.ui.nodes['agent.options']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.09 07:00:00                  #
# ================================================== #

import mimetypes
import os.path
import shutil
import ssl
from urllib.request import Request, urlopen
from PySide6.QtCore import Slot

from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    @Slot()
    def run(self):
        for item in self.cmds:
            try:
                if item["cmd"] in self.plugin.allowed_cmds and self.plugin.is_cmd_allowed(item["cmd"]):

                    # save file
                    if item["cmd"] == "save_file":
                        self.cmd_save_file(item)

                    # append to file
                    elif item["cmd"] == "append_file":
                        self.cmd_append_file(item)

                    # read file
                    elif item["cmd"] == "read_file":
                        self.cmd_read_file(item)

                    # query file
                    elif item["cmd"] == "query_file":
                        self.cmd_query_file(item)

                    # delete file
                    elif item["cmd"] == "delete_file":
                        self.cmd_delete_file(item)

                    # list files
                    elif item["cmd"] == "list_dir":
                        self.cmd_list_dir(item)

                    # mkdir
                    elif item["cmd"] == "mkdir":
                        self.cmd_mkdir(item)

                    # rmdir
                    elif item["cmd"] == "rmdir":
                        self.cmd_rmdir(item)

                    # download
                    elif item["cmd"] == "download_file":
                        self.cmd_download_file(item)

                    # copy file
                    elif item["cmd"] == "copy_file":
                        self.cmd_copy_file(item)

                    # copy dir
                    elif item["cmd"] == "copy_dir":
                        self.cmd_copy_dir(item)

                    # move
                    elif item["cmd"] == "move":
                        self.cmd_move(item)

                    # is dir
                    elif item["cmd"] == "is_dir":
                        self.cmd_is_dir(item)

                    # is file
                    elif item["cmd"] == "is_file":
                        self.cmd_is_file(item)

                    # file exists
                    elif item["cmd"] == "file_exists":
                        self.cmd_file_exists(item)

                    # file size
                    elif item["cmd"] == "file_size":
                        self.cmd_file_size(item)

                    # file info
                    elif item["cmd"] == "file_info":
                        self.cmd_file_info(item)

                    # cwd
                    elif item["cmd"] == "cwd":
                        self.cmd_cwd(item)

                    # get file as attachment
                    elif item["cmd"] == "send_file":
                        self.cmd_send_file(item)

                    # index file or directory
                    elif item["cmd"] == "file_index":
                        self.cmd_file_index(item)

            except Exception as e:
                self.response(
                    {
                        "request": item,
                        "result": "Error: {}".format(e),
                    },
                    self.get_extra_data(),
                )
                self.error(e)
                self.log("Error: {}".format(e))

        if self.msg is not None:
            self.status(self.msg)

    def prepare_request(self, item) -> dict:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        request = {"cmd": item["cmd"]}  # prepare request item for result
        return request

    def get_extra_data(self) -> dict:
        """
        Return extra data for response

        :return: extra data
        """
        return {
            "post_update": ["file_explorer"],  # update file explorer after processing
        }

    def cmd_save_file(self, item: dict):
        """
        Save file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Saving file: {}".format(item["params"]['filename'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['filename'])
            data = item["params"]['data']
            with open(path, 'w', encoding="utf-8") as file:
                file.write(data)
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("File saved: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_append_file(self, item: dict):
        """
        Append to file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Appending file: {}".format(item["params"]['filename'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['filename'])
            data = item["params"]['data']
            with open(path, 'a', encoding="utf-8") as file:
                file.write(data)
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("File appended: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_read_file(self, item: dict):
        """
        Read file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Reading file: {}".format(item["params"]['filename'])
            self.log(self.msg)
            path = item["params"]['filename']
            paths = []
            if isinstance(path, list):
                paths = path
            elif isinstance(path, str):
                paths = [path]
            data, context = self.read_files(paths)
            context_str = None
            if context:
                context_str = "\n\n".join(context)
            response = {
                "request": request,
                "result": data,
            }
            if context_str:
                response["context"] = context_str
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_query_file(self, item: dict):
        """
        Query file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        response = {}
        query = None
        try:
            self.msg = "Reading path: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if "query" in item["params"] and item["params"]["query"]:
                query = item["params"]["query"]

            # check if file exists
            if os.path.exists(path):
                if query is not None:
                    # query file using temp index (created on the fly)
                    self.log("Querying file: {}".format(path))

                    # get tmp query model
                    model = self.plugin.window.core.models.from_defaults()
                    tmp_model = self.plugin.get_option_value("model_tmp_query")
                    if self.plugin.window.core.models.has(tmp_model):
                        model = self.plugin.window.core.models.get(tmp_model)

                    answer = self.plugin.window.core.idx.chat.query_file(
                        ctx=self.ctx,
                        path=path,
                        query=query,
                        model=model,
                    )
                    self.log("Response from temporary in-memory index: {}".format(answer))
                    if answer:
                        response = {
                            "request": request,
                            "result": answer,
                            "context": "From: " + os.path.basename(path) + ":\n--------------------------------\n" + answer,
                            # add additional context
                        }

                # + auto-index file to main index using Llama-index
                if self.plugin.get_option_value("auto_index"):
                    idx_name = self.plugin.get_option_value("idx")
                    self.plugin.window.core.idx.index_files(
                        idx_name,
                        path,
                    )
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_delete_file(self,item: dict):
        """
        Delete file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Deleting file: {}".format(item["params"]['filename'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['filename'])
            if os.path.exists(path):
                os.remove(path)
                response = {"request": request, "result": "OK"}
                self.log("File deleted: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_list_dir(self, item: dict):
        """
        List directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Listing directory: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                files = os.listdir(path)
                response = {
                    "request": request,
                    "result": files,
                }
                self.log("Files listed: {}".format(path))
                self.log("Result: {}".format(files))
            else:
                response = {
                    "request": request,
                    "result": "Directory not found",
                }
                self.log("Directory not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_mkdir(self, item: dict):
        """
        Make directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Creating directory: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if not os.path.exists(path):
                os.makedirs(path)
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("Directory created: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "Directory already exists",
                }
                self.log("Directory already exists: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_rmdir(self, item: dict):
        """
        Remove directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Deleting directory: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                shutil.rmtree(path)
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("Directory deleted: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "Directory not found",
                }
                self.log("Directory not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_download_file(self, item: dict):
        """
        Download file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            dst = self.prepare_path(item["params"]['dst'])
            self.msg = "Downloading file: {} into {}".format(item["params"]['src'], dst)
            self.log(self.msg)
            # Check if src is URL
            if item["params"]['src'].startswith("http"):
                src = item["params"]['src']
                # Download file from URL
                req = Request(
                    url=src,
                    headers={'User-Agent': 'Mozilla/5.0'},
                )
                context = ssl.create_default_context()
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                with urlopen(
                        req,
                        context=context,
                        timeout=5) as response, \
                        open(dst, 'wb') as out_file:
                    shutil.copyfileobj(response, out_file)

                size = os.path.getsize(dst)
            else:
                # Handle local file paths
                src = os.path.join(
                    self.plugin.window.core.config.get_user_dir('data'),
                    item["params"]['src'],
                )

                # Copy local file
                with open(src, 'rb') as in_file, open(dst, 'wb') as out_file:
                    shutil.copyfileobj(in_file, out_file)
                size = os.path.getsize(dst)

            # handle result
            response = {
                "request": request,
                "result": {
                    "result": "OK",
                    "size_bytes": size,
                    "size_human": self.get_human_readable_size(size),
                }
            }
            self.log("File downloaded: {} into {}".format(src, dst))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_copy_file(self, item: dict):
        """
        Copy file

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Copying file: {} into {}".format(item["params"]['src'], item["params"]['dst'])
            self.log(self.msg)
            dst = self.prepare_path(item["params"]['dst'])
            src = self.prepare_path(item["params"]['src'])
            shutil.copyfile(src, dst)
            response = {
                "request": request,
                "result": "OK",
            }
            self.log("File copied: {} into {}".format(src, dst))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_copy_dir(self, item: dict):
        """
        Copy directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Copying directory: {} into {}".format(item["params"]['src'],
                                                              item["params"]['dst'])
            self.log(self.msg)
            dst = self.prepare_path(item["params"]['dst'])
            src = self.prepare_path(item["params"]['src'])
            shutil.copytree(src, dst)
            response = {
                "request": request,
                "result": "OK",
            }
            self.log("Directory copied: {} into {}".format(src, dst))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_move(self, item: dict):
        """
        Move file or directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Moving: {} into {}".format(item["params"]['src'], item["params"]['dst'])
            self.log(self.msg)
            dst = self.prepare_path(item["params"]['dst'])
            src = self.prepare_path(item["params"]['src'])
            shutil.move(src, dst)
            response = {
                "request": request,
                "result": "OK",
            }
            self.log("Moved: {} into {}".format(src, dst))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_is_dir(self, item: dict):
        """
        Check if directory exists

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Checking if directory exists: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.isdir(path):
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("Directory exists: {}".format(path))
            else:
                response = {"request": request, "result": "Directory not found"}
                self.log("Directory not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_is_file(self, item: dict):
        """
        Check if file exists

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Checking if file exists: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.isfile(path):
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("File exists: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_file_exists(self, item: dict):
        """
        Check if file exists

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Checking if path exists: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                response = {
                    "request": request,
                    "result": "OK",
                }
                self.log("Path exists: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "File or directory not found",
                }
                self.log("Path not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_file_size(self, item: dict):
        """
        Check file size

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Checking file size: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                size = os.path.getsize(path)
                response = {
                    "request": request,
                    "result": {
                        'size_bytes': size,
                        'size_human': self.plugin.human_readable_size(size),
                    },
                }
                self.log("File size: {}".format(size))
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_file_info(self, item: dict):
        """
        Check file info

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Checking file info: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                size = os.path.getsize(path)
                data = {
                    "size": size,
                    "size_human": self.get_human_readable_size(size),
                    'mime_type': mimetypes.guess_type(path)[0] or 'application/octet-stream',
                    "last_access": os.path.getatime(path),
                    "last_modification": os.path.getmtime(path),
                    "creation_time": os.path.getctime(path),
                    "is_dir": os.path.isdir(path),
                    "is_file": os.path.isfile(path),
                    "is_link": os.path.islink(path),
                    "is_mount": os.path.ismount(path),
                    'stat': os.stat(path),
                }
                response = {
                    "request": request,
                    "result": data,
                }
                self.log("File info: {}".format(data))
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_cwd(self, item: dict):
        """
        Get current working directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Getting CWD: {}".format(self.plugin.window.core.config.get_user_dir('data'))
            self.log(self.msg)
            response = {
                "request": request,
                "result": self.plugin.window.core.config.get_user_dir('data'),
            }
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_send_file(self, item: dict):
        """
        Get file as attachment

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Adding attachment: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                # make attachment
                mode = self.plugin.window.core.config.get('mode')
                title = os.path.basename(path)
                self.plugin.window.core.attachments.new(mode, title, path, False)
                self.plugin.window.core.attachments.save()
                self.plugin.window.controller.attachment.update()
                response = {
                    "request": request,
                    "result": "Sending attachment: {}".format(title),
                }
                self.log("Added attachment: {}".format(path))
            else:
                response = {
                    "request": request,
                    "result": "File not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def cmd_file_index(self, item: dict):
        """
        Index file or directory

        :param item: item with parameters
        """
        request = self.prepare_request(item)
        try:
            self.msg = "Indexing path: {}".format(item["params"]['path'])
            self.log(self.msg)
            path = self.prepare_path(item["params"]['path'])
            if os.path.exists(path):
                idx_name = self.plugin.get_option_value("idx")
                # index path using Llama-index
                files, errors = self.plugin.window.core.idx.index_files(
                    idx_name,
                    path,
                )
                data = {
                    'num_indexed': len(files),
                    'index_name': idx_name,
                    'errors': errors,
                    'path': path,
                }
                response = {
                    "request": request,
                    "result": data,
                }
            else:
                response = {
                    "request": request,
                    "result": "File or directory not found",
                }
                self.log("File not found: {}".format(path))
        except Exception as e:
            response = {
                "request": request,
                "result": "Error: {}".format(e),
            }
            self.error(e)
            self.log("Error: {}".format(e))
        self.response(response, self.get_extra_data())

    def get_human_readable_size(self, size, decimal_places=2):
        """
        Return a human-readable file size.

        :param size: file size in bytes
        :param decimal_places: number of decimal places
        :return: human-readable file size
        """
        for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
            if size < 1024.0:
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f} {unit}"

    def is_absolute_path(self, path: str) -> bool:
        """
        Check if path is absolute

        :param path: path to check
        :return: True if absolute
        """
        return os.path.isabs(path)

    def prepare_path(self, path: str) -> str:
        """
        Prepare path

        :param path: path to prepare
        :return: prepared path
        """
        if self.is_absolute_path(path):
            return path
        else:
            return os.path.join(
                self.plugin.window.core.config.get_user_dir('data'),
                path,
            )

    def read_files(self, paths: list) -> (dict, str):
        """
        Read files from directory

        :param paths: list of paths
        :return: response data and context
        """
        data = []
        context = []
        for path in paths:
            path = self.prepare_path(path)
            if os.path.exists(path):
                # + auto-index file using Llama-index
                if self.plugin.get_option_value("auto_index") \
                        or self.plugin.get_option_value("only_index"):
                    idx_name = self.plugin.get_option_value("idx")
                    files, errors = self.plugin.window.core.idx.index_files(
                        idx_name,
                        path,
                    )
                    # if only index, return response and continue
                    if self.plugin.get_option_value("only_index"):
                        data.append({
                            'num_indexed': len(files),
                            'index_name': idx_name,
                            'errors': errors,
                            'path': path,
                        })
                        self.log("File read (index only): {}".format(path))
                        return data, context

                # read file as text
                content = self.plugin.read_as_text(
                    path,
                    use_loaders=self.plugin.get_option_value("use_loaders"),
                )
                context.append(os.path.basename(path) + ":\n--------------------------------\n" + content)
                self.log("File read: {}".format(path))
            else:
                self.log("File not found: {}".format(path))
                data.append("File not found: {}".format(path))

        return data, context



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.15 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QSplitter, QWidget, QSizePolicy

from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.editor import EditorDialog
from pygpt_net.utils import trans


class Preset(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Preset, self).__init__(window, *args, **kwargs)
        """
        Preset editor dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "preset"
        self.dialog_id = "preset.presets"

    def setup(self):
        """Setup preset editor dialog"""
        self.window.ui.nodes['preset.btn.current'] = QPushButton(trans("dialog.preset.btn.current"))
        self.window.ui.nodes['preset.btn.save'] = QPushButton(trans("dialog.preset.btn.save"))
        self.window.ui.nodes['preset.btn.current'].clicked.connect(
            lambda: self.window.controller.presets.editor.from_current())
        self.window.ui.nodes['preset.btn.save'].clicked.connect(
            lambda: self.window.controller.presets.editor.save())

        self.window.ui.nodes['preset.btn.current'].setAutoDefault(False)
        self.window.ui.nodes['preset.btn.save'].setAutoDefault(True)

        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['preset.btn.current'])
        footer.addWidget(self.window.ui.nodes['preset.btn.save'])

        # fields
        self.window.ui.paths[self.id] = QLabel(str(self.window.core.config.path))

        # get option fields config
        fields = self.window.controller.presets.editor.get_options()

        # build settings widgets
        widgets = self.build_widgets(self.id, fields)  # from base config dialog

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[self.id][key] = widgets[key]

        # btn: add function
        self.window.ui.config[self.id]['tool.function'].add_btn.setText(trans('assistant.func.add'))

        # apply widgets to layouts
        options = {}
        for key in widgets:
            if fields[key]["type"] == 'text' or fields[key]["type"] == 'int' or fields[key]["type"] == 'float':
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'textarea':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool':
                widgets[key].setMaximumHeight(38)
                options[key] = self.add_raw_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'dict':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'combo':
                options[key] = self.add_option(widgets[key], fields[key])

        self.window.ui.nodes['preset.tool.function.label'].setVisible(False)  # hide label

        rows = QVBoxLayout()

        ignore_keys = ["chat", "completion", "img", "vision", "llama_index", "langchain", "agent", "prompt", "tool.function"]

        options["chat"].setContentsMargins(0, 0, 0, 0)
        options["completion"].setContentsMargins(0, 0, 0, 0)
        options["img"].setContentsMargins(0, 0, 0, 0)
        options["vision"].setContentsMargins(0, 0, 0, 0)
        options["langchain"].setContentsMargins(0, 0, 0, 0)
        options["llama_index"].setContentsMargins(0, 0, 0, 0)
        options["agent"].setContentsMargins(0, 0, 0, 0)
        options["prompt"].setContentsMargins(0, 0, 0, 0)
        options["tool.function"].setContentsMargins(0, 0, 0, 0)

        rows_mode = QVBoxLayout()
        rows_mode.addLayout(options["chat"])
        rows_mode.addLayout(options["completion"])
        rows_mode.addLayout(options["img"])
        rows_mode.addLayout(options["vision"])
        # rows_mode.addLayout(options["assistant"])
        rows_mode.addLayout(options["langchain"])
        rows_mode.addLayout(options["llama_index"])
        rows_mode.addLayout(options["agent"])

        rows_mode.addStretch()
        rows_mode.setContentsMargins(0, 0, 0, 0)

        widget_mode = QWidget()
        widget_mode.setLayout(rows_mode)
        widget_mode.setContentsMargins(0, 0, 0, 0)

        widget_prompt = QWidget()
        widget_prompt.setLayout(options["prompt"])
        widget_prompt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        widget_tools = QWidget()
        widget_tools.setLayout(options["tool.function"])
        widget_tools.setMinimumWidth(400)

        # append widgets options layouts to rows
        for key in options:
            if key in ignore_keys:
                continue
            options[key].setContentsMargins(0, 0, 0, 0)
            rows.addLayout(options[key])

        rows.setContentsMargins(0, 0, 0, 0)

        rows.addStretch()
        widget_base = QWidget()
        widget_base.setLayout(rows)

        # set max width to options
        widget_base.setMinimumWidth(400)
        widget_base.setMaximumWidth(450)
        widget_mode.setMaximumWidth(300)

        main = QHBoxLayout()
        main.addWidget(widget_base)

        main.addWidget(widget_tools)
        main.addWidget(widget_mode)
        main.setContentsMargins(0, 0, 0, 0)

        widget_main = QWidget()
        widget_main.setLayout(main)

        self.window.ui.splitters['editor.presets'] = QSplitter(Qt.Vertical)
        self.window.ui.splitters['editor.presets'].addWidget(widget_main)
        self.window.ui.splitters['editor.presets'].addWidget(widget_prompt)
        self.window.ui.splitters['editor.presets'].setStretchFactor(0, 1)
        self.window.ui.splitters['editor.presets'].setStretchFactor(1, 2)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.splitters['editor.presets'])
        layout.addLayout(footer)

        self.window.ui.dialog['editor.' + self.dialog_id] = EditorDialog(self.window, self.dialog_id)
        self.window.ui.dialog['editor.' + self.dialog_id].setLayout(layout)
        self.window.ui.dialog['editor.' + self.dialog_id].setWindowTitle(trans('dialog.preset'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.12 04:00:00                  #
# ================================================== #

from packaging.version import parse as parse_version, Version


class Patch:
    def __init__(self, window=None):
        self.window = window

    def execute(self, version: Version) -> bool:
        """
        Migrate to current app version

        :param version: current app version
        :return: True if migrated
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.27 22:00:00                  #
# ================================================== #

import os.path
from llama_index.core.indices.service_context import ServiceContext
from llama_index.llms.openai import OpenAI

from pygpt_net.item.model import ModelItem


class Llm:
    def __init__(self, window=None):
        """
        Index LLM core

        :param window: Window instance
        """
        self.window = window
        self.indexes = {}

    def init(self):
        """Init env vars"""
        os.environ['OPENAI_API_KEY'] = str(self.window.core.config.get('api_key'))
        os.environ['OPENAI_ORGANIZATION'] = str(self.window.core.config.get('organization_key'))

    def get(self, model: ModelItem = None):
        """
        Get LLM provider

        :param model: Model item
        :return: LLM
        """
        llm = None
        if model is not None:
            if 'provider' in model.llama_index:
                provider = model.llama_index['provider']
                if provider in self.window.core.llm.llms:
                    try:
                        # init
                        self.window.core.llm.llms[provider].init(
                            window=self.window,
                            model=model,
                            mode="llama_index",
                            sub_mode="",
                        )
                        # get llama llm instance
                        llm = self.window.core.llm.llms[provider].llama(
                            window=self.window,
                            model=model,
                        )
                    except Exception as e:
                        print(e)

        # default
        if llm is None:
            self.init()  # init env vars
            llm = OpenAI(
                temperature=0.0,
                model="gpt-3.5-turbo",
            )
        return llm

    def get_service_context(self, model: ModelItem = None) -> ServiceContext:
        """
        Get service context

        :param model: Model item
        :return: Service context
        """
        llm = self.get(model=model)
        if llm is None:
            return ServiceContext.from_defaults()
        return ServiceContext.from_defaults(llm=llm)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.31 04:00:00                  #
# ================================================== #

class Info:
    def __init__(self, window=None):
        """
        Info core

        :param window main window
        """
        self.window = window

        # prepare info ids
        self.ids = ['about', 'changelog']
        self.active = {}

        # prepare active
        for id in self.ids:
            self.active[id] = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.11 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget, QTabWidget, QFrame, \
    QSplitter, QSizePolicy

from pygpt_net.plugin.base import BasePlugin
from pygpt_net.ui.widget.dialog.settings_plugin import PluginSettingsDialog
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.element.labels import UrlLabel, HelpLabel
from pygpt_net.ui.widget.lists.plugin import PluginList
from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.cmd import OptionCmd
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.ui.widget.option.input import OptionInput, PasswordInput
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.utils import trans


class Plugins:
    def __init__(self, window=None):
        """
        Plugin settings dialog

        :param window: Window instance
        """
        self.window = window
        self.dialog_id = "plugin_settings"
        self.max_list_width = 250

    def setup(self, idx=None):
        """
        Setup plugin settings dialog

        :param idx: current plugin tab index
        """
        self.window.ui.nodes['plugin.settings.btn.defaults.user'] = \
            QPushButton(trans("dialog.plugin.settings.btn.defaults.user"))
        self.window.ui.nodes['plugin.settings.btn.defaults.app'] = \
            QPushButton(trans("dialog.plugin.settings.btn.defaults.app"))
        self.window.ui.nodes['plugin.settings.btn.save'] = \
            QPushButton(trans("dialog.plugin.settings.btn.save"))

        self.window.ui.nodes['plugin.settings.btn.defaults.user'].clicked.connect(
            lambda: self.window.controller.plugins.settings.load_defaults_user())
        self.window.ui.nodes['plugin.settings.btn.defaults.app'].clicked.connect(
            lambda: self.window.controller.plugins.settings.load_defaults_app())
        self.window.ui.nodes['plugin.settings.btn.save'].clicked.connect(
            lambda: self.window.controller.plugins.settings.save())

        # set enter key to save button
        self.window.ui.nodes['plugin.settings.btn.defaults.user'].setAutoDefault(False)
        self.window.ui.nodes['plugin.settings.btn.defaults.app'].setAutoDefault(False)
        self.window.ui.nodes['plugin.settings.btn.save'].setAutoDefault(True)

        # footer buttons
        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['plugin.settings.btn.defaults.user'])
        footer.addWidget(self.window.ui.nodes['plugin.settings.btn.defaults.app'])
        footer.addWidget(self.window.ui.nodes['plugin.settings.btn.save'])

        # plugins tabs
        self.window.ui.tabs['plugin.settings'] = QTabWidget()
        self.window.ui.tabs['plugin.settings.tabs'] = {}

        # build plugin settings tabs
        for id in self.window.core.plugins.plugins:
            content_tabs = {}
            scroll_tabs = {}

            plugin = self.window.core.plugins.plugins[id]
            parent_id = "plugin." + id

            # create plugin options entry if not exists
            if parent_id not in self.window.ui.config:
                self.window.ui.config[parent_id] = {}

            # get plugin options
            options = plugin.setup()

            # extract tab ids, general is default
            tab_ids = self.extract_option_tabs(options)
            for tab_id in tab_ids:
                content_tabs[tab_id] = QVBoxLayout()
                scroll_tabs[tab_id] = QScrollArea()
                scroll_tabs[tab_id].setWidgetResizable(True)

            widgets = self.build_widgets(plugin, options)
            advanced_keys = []
            for key in options:
                if 'advanced' in options[key] and options[key]['advanced']:
                    advanced_keys.append(key)

            # apply settings widgets
            for key in widgets:
                self.window.ui.config[parent_id][key] = widgets[key]

            # append URLs at the beginning
            if len(plugin.urls) > 0:
                tab_id = "general"
                urls_widget = self.add_urls(plugin.urls)
                content_tabs[tab_id].addWidget(urls_widget)

            for key in widgets:
                if key in advanced_keys:  # hide advanced options
                    continue

                tab_id = "general"
                if 'tab' in options[key]:
                    tab = options[key]['tab']
                    if tab is not None and tab != "":
                        tab_id = tab

                content_tabs[tab_id].addLayout(
                    self.add_option(
                        plugin,
                        widgets[key],
                        options[key],
                    )
                )  # add to scroll

            # append advanced options at the end
            if len(advanced_keys) > 0:
                groups = {}
                group_id = 'plugin.settings.advanced' + '.' + id
                for key in widgets:
                    if key not in advanced_keys:  # ignore non-advanced options
                        continue

                    tab_id = "general"
                    if 'tab' in options[key]:
                        tab = options[key]['tab']
                        if tab is not None and tab != "":
                            tab_id = tab
                    if tab_id not in groups:
                        full_id = group_id + '.' + tab_id
                        groups[tab_id] = CollapsedGroup(self.window, full_id, None, False, None)
                        groups[tab_id].box.setText(trans('settings.advanced.collapse'))

                    option = self.add_option(plugin, widgets[key], options[key])  # build option
                    groups[tab_id].add_layout(option)  # add option to group

                # add advanced options group to scrolls
                for tab_id in groups:
                    full_id = group_id + '.' + tab_id
                    content_tabs[tab_id].addWidget(groups[tab_id])
                    self.window.ui.groups[full_id] = groups[tab_id]

            # add stretch to the end of every option tab
            for tab_id in content_tabs:
                content_tabs[tab_id].addStretch()

            # set description, translate if localization is enabled
            name_txt = plugin.name
            desc_key = 'plugin.settings.' + id + '.desc'
            desc_txt = plugin.description
            if plugin.use_locale:
                domain = 'plugin.' + plugin.id
                name_txt = trans('plugin.name', False, domain)
                desc_txt = trans('plugin.description', False, domain)

            self.window.ui.nodes[desc_key] = QLabel(desc_txt)
            self.window.ui.nodes[desc_key].setWordWrap(True)
            self.window.ui.nodes[desc_key].setAlignment(Qt.AlignCenter)
            self.window.ui.nodes[desc_key].setStyleSheet("font-weight: bold;")

            line = self.add_line()

            # tabs or no tabs
            if len(content_tabs) > 1:
                # add tabs
                tab_widget = QTabWidget()

                # sort to make general tab first if exists
                if "general" in content_tabs:
                    content_tabs = {"general": content_tabs.pop("general")} | content_tabs

                for tab_id in content_tabs:
                    tab_name = tab_id
                    if tab_id in plugin.tabs:
                        tab_name = plugin.tabs[tab_id]
                        if tab_id == "general":
                            tab_name = trans("plugin.tab.general")
                    scroll_widget = QWidget()
                    scroll_widget.setLayout(content_tabs[tab_id])
                    scroll_tabs[tab_id].setWidget(scroll_widget)
                    tab_widget.addTab(scroll_tabs[tab_id], tab_name)

                area = QVBoxLayout()
                area.addWidget(self.window.ui.nodes[desc_key])
                area.addWidget(line)
                area.addWidget(tab_widget)
            else:
                # scroll widget
                scroll_widget = QWidget()
                scroll_widget.setLayout(content_tabs["general"])
                scroll_tabs["general"].setWidget(scroll_widget)

                area = QVBoxLayout()
                area.addWidget(self.window.ui.nodes[desc_key])
                area.addWidget(line)
                area.addWidget(scroll_tabs["general"])

            area_widget = QWidget()
            area_widget.setLayout(area)

            # append to tab
            self.window.ui.tabs['plugin.settings'].addTab(area_widget, name_txt)

        # -----------------------

        self.window.ui.tabs['plugin.settings'].currentChanged.connect(
            lambda: self.window.controller.plugins.set_by_tab(
                self.window.ui.tabs['plugin.settings'].currentIndex())
        )

        data = {}
        for plugin_id in self.window.core.plugins.plugins:
            plugin = self.window.core.plugins.plugins[plugin_id]
            data[plugin_id] = plugin

        # plugins list
        id = 'plugin.list'
        self.window.ui.nodes[id] = PluginList(self.window, id)
        self.window.ui.models[id] = self.create_model(self.window)
        self.window.ui.nodes[id].setModel(self.window.ui.models[id])

        # update plugins list
        self.update_list(id, data)

        # set max width to list
        self.window.ui.nodes[id].setMinimumWidth(self.max_list_width)

        # splitter
        self.window.ui.splitters['dialog.plugins'] = QSplitter(Qt.Horizontal)
        self.window.ui.splitters['dialog.plugins'].addWidget(self.window.ui.nodes[id])  # list
        self.window.ui.splitters['dialog.plugins'].addWidget(self.window.ui.tabs['plugin.settings'])  # tabs
        self.window.ui.splitters['dialog.plugins'].setStretchFactor(0, 2)
        self.window.ui.splitters['dialog.plugins'].setStretchFactor(1, 5)
        self.window.ui.splitters['dialog.plugins'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.window.ui.splitters['dialog.plugins'])

        self.window.ui.nodes['plugin.settings.cmd.footer'] = HelpLabel(trans('cmd.tip'))
        self.window.ui.nodes['plugin.settings.cmd.footer'].setAlignment(Qt.AlignCenter)
        self.window.ui.nodes['plugin.settings.cmd.footer'].setWordWrap(False)

        layout = QVBoxLayout()
        layout.addLayout(main_layout)  # list + plugins tabs
        layout.addLayout(footer)  # bottom buttons (save, defaults)
        layout.addWidget(self.window.ui.nodes['plugin.settings.cmd.footer'])  # command tip

        self.window.ui.dialog[self.dialog_id] = PluginSettingsDialog(self.window, self.dialog_id)
        self.window.ui.dialog[self.dialog_id].setLayout(layout)
        self.window.ui.dialog[self.dialog_id].setWindowTitle(trans('dialog.plugin_settings'))

        # restore current opened tab if idx is set
        if idx is not None:
            try:
                self.window.ui.tabs['plugin.settings'].setCurrentIndex(idx)
                self.window.controller.plugins.set_by_tab(idx)
            except Exception as e:
                print('Failed restore plugin settings tab: {}'.format(idx))
        else:
            self.window.controller.plugins.set_by_tab(0)

    def extract_option_tabs(self, options: dict) -> list:
        """
        Get keys for option tabs

        :param options: plugin options
        :return: list with keys
        """
        keys = []
        is_default = False
        for key in options:
            option = options[key]
            if 'tab' in option:
                tab = option['tab']
                if tab == "" or tab is None:
                    is_default = True
                if tab not in keys:
                    keys.append(tab)
            else:
                is_default = True

        # add default general tab if not exists
        if len(keys) == 0 or (is_default and "general" not in keys):
            keys.append("general")
        return keys

    def build_widgets(self, plugin: BasePlugin, options: dict) -> dict:
        """
        Build settings options widgets

        :param plugin: plugin instance
        :param options: plugin options
        :return: dict of widgets
        """
        id = plugin.id
        parent = "plugin." + id  # parent id for plugins is in format: plugin.<plugin_id>
        widgets = {}

        for key in options:
            option = options[key]
            # create widget by option type
            if option['type'] == 'text' or option['type'] == 'int' or option['type'] == 'float':
                if 'slider' in option and option['slider'] \
                        and (option['type'] == 'int' or option['type'] == 'float'):
                    widgets[key] = OptionSlider(self.window, parent, key, option)  # slider + text input
                else:
                    if 'secret' in option and option['secret']:
                        widgets[key] = PasswordInput(self.window, parent, key, option)  # password input
                    else:
                        widgets[key] = OptionInput(self.window, parent, key, option)  # text input
            elif option['type'] == 'textarea':
                widgets[key] = OptionTextarea(self.window, parent, key, option)  # textarea
            elif option['type'] == 'bool':
                widgets[key] = OptionCheckbox(self.window, parent, key, option)  # checkbox
            elif option['type'] == 'dict':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionDict(self.window, parent, key, option)  # dictionary
            elif option['type'] == 'combo':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionCombo(self.window, parent, key, option)  # combobox
            elif option['type'] == 'cmd':
                widgets[key] = OptionCmd(self.window, parent, key, option)  # command config

        return widgets

    def add_line(self) -> QFrame:
        """
        Make separator line

        :return: separator line
        """
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        return line

    def add_urls(self, urls: dict) -> QWidget:
        """
        Add clickable urls to list

        :param urls: urls dict
        :return: QWidget
        """
        layout = QVBoxLayout()
        if isinstance(urls, dict):
            for name in urls:
                url = urls[name]
                label = UrlLabel(name, url)
                layout.addWidget(label)
        elif isinstance(urls, list):
            for url in urls:
                label = UrlLabel("", url)
                layout.addWidget(label)
        elif isinstance(urls, str):
            label = UrlLabel("", urls)
            layout.addWidget(label)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def add_option(self, plugin: BasePlugin, widget: QWidget, option: dict) -> QVBoxLayout:
        """
        Append option widget to layout

        :param plugin: plugin instance
        :param widget: widget instance
        :param option: option dict
        :return: QVBoxLayout
        """
        one_column_types = ['textarea', 'dict', 'bool', 'cmd']
        no_label_types = ['bool', 'cmd']
        allow_locale = True

        key = option['id']
        label_key = 'plugin.' + plugin.id + '.' + key + '.label'
        desc_key = 'plugin.' + plugin.id + '.' + key + '.desc'

        # get option label and description
        txt_title = option['label']
        txt_desc = option['description']
        txt_tooltip = option['tooltip']

        if "locale" in option and not option["locale"]:
            allow_locale = False

        # translate if localization is enabled
        if plugin.use_locale and allow_locale:
            domain = 'plugin.' + plugin.id
            txt_title = trans(key + '.label', False, domain)
            txt_desc = trans(key + '.description', False, domain)
            txt_tooltip = trans(key + '.tooltip', False, domain)

            # if empty tooltip then use description
            if txt_tooltip == key + '.tooltip':
                txt_tooltip = txt_desc

        widget.setToolTip(txt_tooltip)

        if option['type'] not in no_label_types:
            self.window.ui.nodes[label_key] = QLabel(txt_title)
            self.window.ui.nodes[label_key].setStyleSheet("font-weight: bold;")

        # 2-columns layout
        if option['type'] not in one_column_types:
            cols = QHBoxLayout()

            if option['type'] not in no_label_types:
                cols.addWidget(self.window.ui.nodes[label_key])  # disable label in bool type

            cols.addWidget(widget)

            cols_widget = QWidget()
            cols_widget.setLayout(cols)
            cols_widget.setMaximumHeight(90)

            self.window.ui.nodes[desc_key] = QLabel(txt_desc)
            self.window.ui.nodes[desc_key].setWordWrap(True)
            self.window.ui.nodes[desc_key].setMaximumHeight(40)
            self.window.ui.nodes[desc_key].setStyleSheet("font-size: 10px;")
            self.window.ui.nodes[desc_key].setToolTip(txt_tooltip)

            layout = QVBoxLayout()
            layout.addWidget(cols_widget)
            layout.addWidget(self.window.ui.nodes[desc_key])
        else:
            # 1-column layout: textarea, dict, cmd fields
            layout = QVBoxLayout()

            if option['type'] not in no_label_types:
                layout.addWidget(self.window.ui.nodes[label_key])
            elif option['type'] == 'bool':
                widget.box.setText(txt_title)  # set checkbox label

            layout.addWidget(widget)

            self.window.ui.nodes[desc_key] = QLabel(txt_desc)
            self.window.ui.nodes[desc_key].setWordWrap(True)
            self.window.ui.nodes[desc_key].setMaximumHeight(40)
            self.window.ui.nodes[desc_key].setStyleSheet("font-size: 10px;")
            self.window.ui.nodes[desc_key].setToolTip(txt_tooltip)

            layout.addWidget(self.window.ui.nodes[desc_key])

        # append URLs at the beginning
        if option['urls'] is not None and len(option['urls']) > 0:
            urls_widget = self.add_urls(option['urls'])
            layout.addWidget(urls_widget)

        line = self.add_line()  # TODO: change name to separator
        layout.addWidget(line)

        return layout

    def add_raw_option(self, option: dict) -> QHBoxLayout:
        """
        Add raw option row

        :param option: Option
        :return: QHBoxLayout
        """
        layout = QHBoxLayout()
        layout.addWidget(option)
        return layout

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update_list(self, id: str, data: dict):
        """
        Update list

        :param id: ID of the list
        :param data: Data to update
        """
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[id].insertRow(i)
            name = self.window.core.plugins.get_name(data[n].id)
            tooltip = self.window.core.plugins.get_desc(data[n].id)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), tooltip, Qt.ToolTipRole)
            i += 1



import base64
from pathlib import Path
from typing import Dict, List, Optional

from llama_index.core.readers.base import BaseReader
from llama_index.core.schema import Document, ImageDocument
from llama_index.core.utils import infer_torch_device


class ImageVisionLLMReader(BaseReader):
    """Image parser.

    Caption image using Blip2 (a multimodal VisionLLM similar to GPT4).

    Based on: https://github.com/run-llama/llama_index/blob/main/llama-index-integrations/readers/llama-index-readers-file/llama_index/readers/file/image_vision_llm/base.py

    """

    def __init__(
        self,
        use_local: bool = False,
        window = None,
        keep_image: bool = False,
        api_prompt: str = "Describe what you see in this image",
        api_model: str = "gpt-4-vision-preview",
        api_tokens: int = 1000,
        local_prompt: str = "Question: describe what you see in this image. Answer:",
    ):
        """
        Init params.

        :param use_local: use local model
        :param window: Window instance
        :param keep_image: keep image in document
        :param api_prompt: API prompt
        :param api_model: API model
        :param api_tokens: API tokens
        :param local_prompt: local prompt
        """
        self._initialized = False
        self._use_local = use_local
        self._api_prompt = api_prompt
        self._api_model = api_model
        self._api_tokens = api_tokens
        self._window = window

        self._parser_config = None
        self._keep_image = keep_image
        self._local_prompt = local_prompt

    def _initialize(self) -> None:
        """Initialize local parser."""
        parser_config = None
        if self._parser_config is None:
            try:
                import sentencepiece  # noqa
                import torch
                from PIL import Image  # noqa
                from transformers import Blip2ForConditionalGeneration, Blip2Processor
            except ImportError:
                raise ImportError(
                    "Please install extra dependencies that are required for "
                    "the ImageVisionLLMReader: "
                    "`pip install torch transformers sentencepiece Pillow`"
                )

            device = infer_torch_device()
            dtype = torch.float16 if torch.cuda.is_available() else torch.float32
            processor = Blip2Processor.from_pretrained("Salesforce/blip2-opt-2.7b")
            model = Blip2ForConditionalGeneration.from_pretrained(
                "Salesforce/blip2-opt-2.7b", torch_dtype=dtype
            )
            parser_config = {
                "processor": processor,
                "model": model,
                "device": device,
                "dtype": dtype,
            }

        self._parser_config = parser_config
        self._initialized = True

    def load_local(self, file: Path, extra_info: Optional[Dict] = None) -> List[Document]:
        """
        Describe image using local model.

        :param file: file path
        :param extra_info: additional metadata
        :return: list of documents
        """
        if not self._initialized:
            self._initialize()

        from llama_index.core.img_utils import img_2_b64
        from PIL import Image

        # load document image
        image = Image.open(file)
        if image.mode != "RGB":
            image = image.convert("RGB")

        # Encode image into base64 string and keep in document
        image_str: Optional[str] = None
        if self._keep_image:
            image_str = img_2_b64(image)

        # Parse image into text
        model = self._parser_config["model"]
        processor = self._parser_config["processor"]

        device = self._parser_config["device"]
        dtype = self._parser_config["dtype"]
        model.to(device)

        # unconditional image captioning
        inputs = processor(image, self._local_prompt, return_tensors="pt").to(device, dtype)
        out = model.generate(**inputs)
        text_str = processor.decode(out[0], skip_special_tokens=True)

        return [
            ImageDocument(
                text=text_str,
                image=image_str,
                image_path=str(file),
                metadata=extra_info or {},
            )
        ]

    def load_api(self, file: Path, extra_info: Optional[Dict] = None) -> List[Document]:
        """
        Describe image using OpenAI API.

        :param file: file path
        :param extra_info: additional metadata
        :return: list of documents
        """
        from llama_index.core.img_utils import img_2_b64
        from PIL import Image

        # Encode image into base64 string and keep in document
        image_str: Optional[str] = None
        if self._keep_image:
            image = Image.open(file)
            if image.mode != "RGB":
                image = image.convert("RGB")
            image_str = img_2_b64(image)

        client = self._window.core.gpt.get_client()
        encoded = self._encode_image(str(file))
        content = [
            {
                "type": "text",
                "text": self._api_prompt
            },
            {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/jpeg;base64,{encoded}",
                }
            }
        ]
        messages = []
        messages.append({"role": "user", "content": content})
        response = client.chat.completions.create(
            messages=messages,
            model=self._api_model,
            max_tokens=self._api_tokens,
        )
        text = response.choices[0].message.content.strip()
        return [
            ImageDocument(
                text=text,
                image=image_str,
                image_path=str(file),
                metadata=extra_info or {},
            )
        ]

    def _encode_image(self, image_path: str) -> str:
        """
        Encode image to base64

        :param image_path: path to image
        :return: base64 encoded image
        """
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def load_data(
        self, file: Path, extra_info: Optional[Dict] = None
    ) -> List[Document]:
        """
        Parse file.

        :param file: file path
        :param extra_info: additional metadata
        :return: list of documents
        """
        if self._use_local:
            is_compiled = self._window.core.config.is_compiled() or self._window.core.platforms.is_snap()
            if is_compiled:
                raise ValueError("Local models are not available in compiled version.")
            return self.load_local(file, extra_info)
        else:
            if self._window is None:
                raise ValueError("Window instance is required to use API model.")
            return self.load_api(file, extra_info)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.11 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex, QSize
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QTreeView, QMenu, QStyledItemDelegate, QComboBox, \
    QCheckBox, QSizePolicy

from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class OptionCmd(QWidget):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Settings dictionary items

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionCmd, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option  # option data

        # TODO: implement cmd type option

        # prepare options
        option_enabled = {}
        option_enabled["type"] = "bool"
        option_enabled["label"] = ""

        if self.option["value"] is not None and "cmd" in self.option["value"]:
            option_enabled["label"] = "Command: " + self.option["value"]["cmd"]
        option_enabled["description"] = "Test description"
        option_enabled["value"] = True

        option_params = {}
        option_params["name"] = "params"
        option_params["type"] = "dict"
        option_params["keys"] = {
            "name": "text",
            "type": "text",
            "description": "text",
            "required": "bool",
        }
        option_params["value"] = []

        if self.option["value"] is not None and "params" in self.option["value"]:
            option_params["value"] = self.option["value"]["params"]

        key_enabled = self.id + ".enabled"
        key_params = self.id + ".params"

        self.enabled = OptionCheckbox(self.window, parent_id, key_enabled, option_enabled)  # enable checkbox
        self.params = OptionDict(self.window, parent_id, key_params, option_params)  # command params

        # layout
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.enabled)
        self.layout.addWidget(self.params)
        self.setLayout(self.layout)

        # update
        self.update()

    def update_item(self, idx, data):
        """
        Update item in params list

        :param idx: Item index
        :param data: Item data
        """
        self.params.update_item(idx, data)

    def update(self):
        """Update widget"""
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.11 01:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class Cmd:
    def __init__(self, window=None):
        """
        Command field handler - >>> NOT IMPLEMENTED YET <<<<

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: dict
    ):
        """
        Apply values to command fields

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        if option["value"] is None:
            option["value"] = {
                "enabled": False,
                "instruction": "",
                "params": [],
            }
        values = list(option["value"]["params"])
        self.window.ui.config[parent_id][key].params.items = values  # replace model data list
        self.window.ui.config[parent_id][key].params.model.updateData(values)  # update model data

    def apply_row(
            self,
            parent_id: str,
            key: str,
            values: dict,
            idx: int
    ):
        """
        Apply data values to command fields

        :param parent_id: Options parent ID
        :param key: Option key
        :param values: dictionary data values
        :param idx: row index
        """
        # remove .params suffix
        if key.endswith(".params"):
            key = key.replace(".params", "")
        self.window.ui.config[parent_id][key].params.update_item(idx, values["params"])

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: dict
    ) -> dict:
        """
        Get command values

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :return: list with dictionary values
        """
        data = {
            "enabled": self.window.ui.config[parent_id][key].enabled.box.isChecked(),
            "params": self.window.ui.config[parent_id][key].params.model.items,
        }
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.04 20:00:00                  #
# ================================================== #

import datetime
import json
import time


class CtxItem:
    def __init__(self, mode=None):
        """
        Context item

        :param mode: Mode (completion, chat, img, vision, langchain, assistant)
        """
        self.id = None
        self.meta_id = None
        self.external_id = None
        self.stream = None
        self.cmds = []
        self.results = []
        self.urls = []
        self.images = []
        self.files = []
        self.attachments = []
        self.reply = False
        self.input = None
        self.output = None
        self.mode = mode
        self.model = None
        self.thread = None
        self.msg_id = None
        self.run_id = None
        self.input_name = None
        self.output_name = None
        self.input_timestamp = None
        self.output_timestamp = None
        self.input_tokens = 0
        self.output_tokens = 0
        self.total_tokens = 0
        self.extra = None
        self.extra_ctx = None
        self.current = False
        self.internal = False
        self.is_vision = False
        self.idx = 0
        self.first = False
        self.tool_calls = []  # API tool calls
        self.index_meta = {}  # llama-index metadata ctx used
        self.doc_ids = []  # document ids

    def add_doc_meta(self, meta: dict):
        """
        Add document meta

        :param meta: document meta data
        """
        self.index_meta = meta
        self.doc_ids.append(meta)

    def set_input(self, input: str | None, name: str = None):
        """
        Set input

        :param input: input text (prompt)
        :param name: input person name
        """
        self.input = input
        self.input_name = name
        self.input_timestamp = int(time.time())

    def set_output(self, output: str | None, name: str = None):
        """
        Set output

        :param output: output text
        :param name: output person name
        """
        self.output = output
        self.output_name = name
        self.output_timestamp = int(time.time())

    def set_tokens(self, input_tokens: int, output_tokens: int):
        """
        Set tokens usage

        :param input_tokens: prompt tokens
        :param output_tokens: output tokens
        """
        self.input_tokens = input_tokens
        self.output_tokens = output_tokens
        self.total_tokens = input_tokens + output_tokens

    def to_dict(self) -> dict:
        """
        Dump context item to dict

        :return: dict
        """
        return {
            "id": self.id,
            "meta_id": self.meta_id,
            "external_id": self.external_id,
            # "stream": self.stream,  #  <-- do not dump stream response object
            "cmds": self.cmds,
            "results": self.results,
            "urls": self.urls,
            "images": self.images,
            "files": self.files,
            "attachments": self.attachments,
            "reply": self.reply,
            "input": self.input,
            "output": self.output,
            "mode": self.mode,
            "model": self.model,
            "thread": self.thread,
            "msg_id": self.msg_id,
            "run_id": self.run_id,
            "input_name": self.input_name,
            "output_name": self.output_name,
            "input_timestamp": self.input_timestamp,
            "output_timestamp": self.output_timestamp,
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "total_tokens": self.total_tokens,
            "extra": self.extra,
            "extra_ctx": self.extra_ctx,
            "current": self.current,
            "internal": self.internal,
            "is_vision": self.is_vision,
            "idx": self.idx,
            "first": self.first,
            "tool_calls": self.tool_calls,
            "index_meta": self.index_meta,
            "doc_ids": self.doc_ids,
        }

    def from_dict(self, data: dict):
        """
        Load context item from dict

        :param data: dict
        """
        self.id = data.get("id", None)
        self.meta_id = data.get("meta_id", None)
        self.external_id = data.get("external_id", None)
        self.stream = data.get("stream", None)
        self.cmds = data.get("cmds", [])
        self.results = data.get("results", [])
        self.urls = data.get("urls", [])
        self.images = data.get("images", [])
        self.files = data.get("files", [])
        self.attachments = data.get("attachments", [])
        self.reply = data.get("reply", False)
        self.input = data.get("input", None)
        self.output = data.get("output", None)
        self.mode = data.get("mode", None)
        self.model = data.get("model", None)
        self.thread = data.get("thread", None)
        self.msg_id = data.get("msg_id", None)
        self.run_id = data.get("run_id", None)
        self.input_name = data.get("input_name", None)
        self.output_name = data.get("output_name", None)
        self.input_timestamp = data.get("input_timestamp", None)
        self.output_timestamp = data.get("output_timestamp", None)
        self.input_tokens = data.get("input_tokens", 0)
        self.output_tokens = data.get("output_tokens", 0)
        self.total_tokens = data.get("total_tokens", 0)
        self.extra = data.get("extra", None)
        self.extra_ctx = data.get("extra_ctx", False)
        self.current = data.get("current", False)
        self.internal = data.get("internal", False)
        self.is_vision = data.get("is_vision", False)
        self.idx = data.get("idx", 0)
        self.first = data.get("first", False)
        self.tool_calls = data.get("tool_calls", [])
        self.index_meta = data.get("index_meta", {})
        self.doc_ids = data.get("doc_ids", [])

    def dump(self) -> str:
        """
        Dump context item to JSON string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()


class CtxMeta:
    def __init__(self, id=None):
        """
        Context meta data

        :param id: Context ID
        """
        self.id = id
        self.external_id = None
        self.uuid = None
        self.name = None
        self.date = datetime.datetime.now().strftime("%Y-%m-%d")
        self.created = int(time.time())
        self.updated = int(time.time())
        self.indexed = None
        self.mode = None
        self.model = None
        self.last_mode = None
        self.last_model = None
        self.thread = None
        self.assistant = None
        self.preset = None
        self.run = None
        self.status = None
        self.extra = None
        self.initialized = False
        self.deleted = False
        self.important = False
        self.archived = False
        self.label = 0  # label color
        self.indexes = {}  # indexes data

    def to_dict(self) -> dict:
        """
        Dump context meta to dict

        :return: dict
        """
        return {
            "id": self.id,
            "external_id": self.external_id,
            "uuid": self.uuid,
            "name": self.name,
            "date": self.date,
            "created": self.created,
            "updated": self.updated,
            "indexed": self.indexed,
            "mode": self.mode,
            "model": self.model,
            "last_mode": self.last_mode,
            "last_model": self.last_model,
            "thread": self.thread,
            "assistant": self.assistant,
            "preset": self.preset,
            "run": self.run,
            "status": self.status,
            "extra": self.extra,
            "initialized": self.initialized,
            "deleted": self.deleted,
            "important": self.important,
            "archived": self.archived,
            "label": self.label,
            "indexes": self.indexes,
        }

    def from_dict(self, data: dict):
        """
        Load context meta from dict

        :param data: dict
        """
        self.id = data.get("id", None)
        self.external_id = data.get("external_id", None)
        self.uuid = data.get("uuid", None)
        self.name = data.get("name", None)
        self.date = data.get("date", None)
        self.created = data.get("created", None)
        self.updated = data.get("updated", None)
        self.indexed = data.get("indexed", None)
        self.mode = data.get("mode", None)
        self.model = data.get("model", None)
        self.last_mode = data.get("last_mode", None)
        self.last_model = data.get("last_model", None)
        self.thread = data.get("thread", None)
        self.assistant = data.get("assistant", None)
        self.preset = data.get("preset", None)
        self.run = data.get("run", None)
        self.status = data.get("status", None)
        self.extra = data.get("extra", None)
        self.initialized = data.get("initialized", False)
        self.deleted = data.get("deleted", False)
        self.important = data.get("important", False)
        self.archived = data.get("archived", False)
        self.label = data.get("label", 0)
        self.indexes = data.get("indexes", {})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.09 07:00:00                  #
# ================================================== #

import json

from PySide6.QtCore import Slot
from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.websearch = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        self.websearch.signals = self.signals  # connect signals

        msg = None
        for item in self.cmds:
            request = {"cmd": item["cmd"]}  # prepare request item for result

            try:
                # cmd: web_search
                if item["cmd"] == "web_search":
                    page = 1
                    if "page" in item["params"]:
                        page = int(item["params"]["page"])
                    prompt = None
                    if "summarize_prompt" in item["params"]:
                        prompt = item["params"]["summarize_prompt"]

                    # search for query
                    result, total_found, current, url = self.websearch.make_query(
                        item["params"]["query"],
                        page,
                        prompt,
                    )
                    msg = "Web search finished: '{}'".format(item["params"]["query"])
                    data = {
                        'content': result,
                        'url': url,
                        'page': current,
                        'total_found': total_found,
                    }
                    if url:
                        self.ctx.urls.append(url)

                    self.response(
                        {
                            "request": request,
                            "result": data,
                        }
                    )

                # cmd: web_url_open
                elif item["cmd"] == "web_url_open":
                    prompt = None
                    if "summarize_prompt" in item["params"]:
                        prompt = item["params"]["summarize_prompt"]
                    url = item["params"]["url"]
                    msg = "Opening Web URL: '{}'".format(item["params"]["url"])

                    # open url
                    result, url = self.websearch.open_url(
                        url,
                        prompt,
                    )
                    data = {
                        'content': result,
                        'url': url,
                    }
                    if url:
                        self.ctx.urls.append(url)

                    self.response(
                        {
                            "request": request,
                            "result": data,
                        }
                    )

                # cmd: web_url_raw
                elif item["cmd"] == "web_url_raw":
                    url = item["params"]["url"]
                    msg = "Opening Web URL: '{}'".format(item["params"]["url"])

                    # open url (raw)
                    result, url = self.websearch.open_url_raw(
                        url,
                    )
                    data = {
                        'content': result,
                        'url': url,
                    }
                    if url:
                        self.ctx.urls.append(url)

                    self.response(
                        {
                            "request": request,
                            "result": data,
                        }
                    )

                # cmd: web_urls
                elif item["cmd"] == "web_urls":
                    page = 1
                    num = 10
                    if "page" in item["params"]:
                        page = int(item["params"]["page"])
                    if "num_links" in item["params"]:
                        num = int(item["params"]["num_links"])
                    if num < 1:
                        num = 1
                    if num > 10:
                        num = 10
                    offset = 1
                    if page > 1:
                        offset = (page - 1) * num + 1

                    # search for URLs
                    urls = self.websearch.google_search(
                        item["params"]["query"],
                        num,
                        offset,
                    )
                    msg = "Web search finished: '{}'".format(item["params"]["query"])
                    data = {
                        'urls': json.dumps(urls),
                        'page': page,
                        'num': num,
                        'offset': offset,
                    }
                    if urls:
                        for url in urls:
                            self.ctx.urls.append(url)

                    self.response(
                        {
                            "request": request,
                            "result": data,
                        }
                    )

                # cmd: web_index
                elif item["cmd"] == "web_index":
                    type = "webpage"  # default
                    args = {}

                    if "type" in item["params"]:
                        type = item["params"]["type"]
                    if "args" in item["params"]:
                        args = item["params"]["args"]

                    url = type
                    if "url" in item["params"]:
                        url = item["params"]["url"]  # from default param
                    if "url" in args:
                        url = args["url"]  # override from args

                    msg = "Indexing URL: '{}'".format(url)
                    idx_name = self.plugin.get_option_value("idx")

                    # show status
                    self.status("Please wait... indexing: {}...".format(url))

                    # index URL via Llama-index
                    num, errors = self.plugin.window.core.idx.index_urls(
                        idx=idx_name,
                        urls=[url],
                        type=type,
                        extra_args=args,
                    )
                    data = {
                        'num_indexed': num,
                        'index': idx_name,
                        'errors': errors,
                        'url': url,
                    }
                    if url and (url.startswith("http://") or url.startswith("https://")):
                            self.ctx.urls.append(url)

                    self.response(
                        {
                            "request": request,
                            "result": data,
                        }
                    )

                # cmd: web_index_query
                elif item["cmd"] == "web_index_query":
                    type = "webpage"  # default
                    args = {}
                    query = None

                    if "type" in item["params"]:
                        type = item["params"]["type"]
                    if "args" in item["params"]:
                        args = item["params"]["args"]

                    url = type
                    if "url" in item["params"]:
                        url = item["params"]["url"]  # from default param
                    if "url" in args:
                        url = args["url"]  # override from args

                    if "query" in item["params"] and item["params"]["query"]:
                        query = item["params"]["query"]

                    response = {
                        "request": request,
                        "result": "No data",
                    }

                    if query is not None:
                        # query file using temp index (created on the fly)
                        self.log("Querying web: {}".format(url))

                        # get tmp query model
                        model = self.plugin.window.core.models.from_defaults()
                        tmp_model = self.plugin.get_option_value("model_tmp_query")
                        if self.plugin.window.core.models.has(tmp_model):
                            model = self.plugin.window.core.models.get(tmp_model)

                        answer = self.plugin.window.core.idx.chat.query_web(
                            ctx=self.ctx,
                            type=type,
                            url=url,
                            args=args,
                            query=query,
                            model=model,
                        )
                        self.log("Response from temporary in-memory index: {}".format(answer))
                        if answer:
                            from_str = type
                            if url:
                                from_str += ", URL: " + url
                            response = {
                                "request": request,
                                "result": answer,
                                "context": "From: " + from_str + ":\n--------------------------------\n" + answer,
                                # add additional context
                            }

                        # + auto-index to main index using Llama-index
                        if self.plugin.get_option_value("auto_index"):
                            msg = "Indexing URL: '{}'".format(url)
                            idx_name = self.plugin.get_option_value("idx")

                            # show status
                            self.status("Please wait... indexing: {}...".format(url))

                            # index URL via Llama-index
                            num, errors = self.plugin.window.core.idx.index_urls(
                                idx=idx_name,
                                urls=[url],
                                type=type,
                                extra_args=args,
                            )

                    # add URL to context
                    if url and (url.startswith("http://") or url.startswith("https://")):
                        self.ctx.urls.append(url)

                    self.response(response)

            except Exception as e:
                self.response(
                    {
                        "request": item,
                        "result": "Error: {}".format(e),
                    }
                )
                self.error(e)
                self.log("Error: {}".format(e))

        if msg is not None:
            self.log(msg)
            self.status(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.30 17:00:00                  #
# ================================================== #

import tiktoken

from pygpt_net.item.ctx import CtxItem

CHAT_MODES = ["chat", "vision", "langchain", "assistant", "llama_index", "agent"]


class Tokens:
    def __init__(self, window=None):
        """
        Tokens core

        :param window: Window instance
        """
        self.window = window

    @staticmethod
    def from_str(string: str, model: str = "gpt-4") -> int:
        """
        Return number of tokens from string

        :param string: string
        :param model: model name
        :return: number of tokens
        """
        if string is None or string == "":
            return 0

        default = "cl100k_base"
        try:
            try:
                if model is not None and model != "":
                    encoding = tiktoken.encoding_for_model(model)
                else:
                    encoding = tiktoken.get_encoding(default)
            except KeyError:
                encoding = tiktoken.get_encoding(default)
            except ValueError:
                return 0

            try:
                return len(encoding.encode(str(string)))
            except Exception as e:
                print("Tokens calc exception", e)
                return 0
        except Exception as e:
            print("Tokens calculation exception:", e)
            return 0

    @staticmethod
    def get_extra(model: str = "gpt-4") -> int:
        """
        Return number of extra tokens

        :param model: model name
        :return: number of tokens
        """
        return 3

    @staticmethod
    def from_prompt(text: str, name: str, model: str = "gpt-4") -> int:
        """
        Return number of tokens from prompt

        :param text: prompt text
        :param name: input name
        :param model: model name
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0

        try:
            num += Tokens.from_str(text, model)
        except Exception as e:
            print("Tokens calc exception", e)

        if name is not None and name != "":
            num += per_message + per_name
        else:
            num += per_message

        return num

    @staticmethod
    def from_text(text: str, model: str = "gpt-4") -> int:
        """
        Return number of tokens from text, without any extra tokens

        :param text: text
        :param model: model name
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0

        if text is None or text == "":
            return 0

        try:
            num += Tokens.from_str(text, model)
        except Exception as e:
            print("Tokens calc exception", e)

        return num

    @staticmethod
    def from_messages(messages: list, model: str = "gpt-4") -> int:
        """
        Return number of tokens from prompt

        :param messages: messages
        :param model: model name
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        for message in messages:
            num += per_message
            for key, value in message.items():
                num += Tokens.from_str(value)
                if key == "name":
                    num += per_name
        num += 3  # every reply is primed with <|start|>assistant<|message|>
        return num

    @staticmethod
    def from_langchain_messages(messages: list, model: str = "gpt-4") -> int:
        """
        Return number of tokens from prompt

        :param messages: messages
        :param model: model name
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        for message in messages:
            num += per_message
            num += Tokens.from_str(message.content)
        num += 3  # every reply is primed with <|start|>assistant<|message|>
        return num

    @staticmethod
    def from_llama_messages(query: str, messages: list, model: str = "gpt-4") -> int:
        """
        Return number of tokens from prompt

        :param query: query
        :param messages: messages
        :param model: model name
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        num += Tokens.from_str(query)
        for message in messages:
            num += per_message
            num += Tokens.from_str(message.content)
        num += 3  # every reply is primed with <|start|>assistant<|message|>
        return num

    @staticmethod
    def from_ctx(ctx: CtxItem, mode: str = "chat", model: str = "gpt-4") -> int:
        """
        Return number of tokens from context ctx

        :param ctx: CtxItem
        :param mode: mode
        :param model: model ID
        :return: number of tokens
        """
        model, per_message, per_name = Tokens.get_config(model)
        num = 0

        if mode in CHAT_MODES:
            # input message
            try:
                num += Tokens.from_str(str(ctx.input), model)
            except Exception as e:
                print("Tokens calc exception", e)

            # output message
            try:
                num += Tokens.from_str(str(ctx.output), model)
            except Exception as e:
                print("Tokens calc exception", e)

            # + fixed tokens
            num += per_message * 2  # input + output
            num += per_name * 2  # input + output
            try:
                num += Tokens.from_str("system", model) * 2  # input + output
            except Exception as e:
                print("Tokens calc exception", e)

            # input name
            if ctx.input_name is not None and ctx.input_name != "":
                name = ctx.input_name
            else:
                name = "user"
            try:
                num += Tokens.from_str(name, model)
            except Exception as e:
                print("Tokens calc exception", e)

            # output name
            if ctx.output_name is not None and ctx.output_name != "":
                name = ctx.output_name
            else:
                name = "assistant"
            try:
                num += Tokens.from_str(name, model)
            except Exception as e:
                print("Tokens calc exception", e)

        # build tmp message if completion mode
        elif mode == "completion":
            message = ""
            # if with names
            if ctx.input_name is not None \
                    and ctx.output_name is not None \
                    and ctx.input_name != "" \
                    and ctx.output_name != "":
                if ctx.input is not None and ctx.input != "":
                    message += "\n" + ctx.input_name + ": " + ctx.input
                if ctx.output is not None and ctx.output != "":
                    message += "\n" + ctx.output_name + ": " + ctx.output
            # if without names
            else:
                if ctx.input is not None and ctx.input != "":
                    message += "\n" + ctx.input
                if ctx.output is not None and ctx.output != "":
                    message += "\n" + ctx.output
            try:
                num += Tokens.from_str(message, model)
            except Exception as e:
                print("Tokens calc exception", e)

        return num

    def get_current(self, input_prompt: str) -> (int, int, int, int, int, int, int, int, int):
        """
        Return current number of used tokens

        :param input_prompt: input prompt
        :return: A tuple of (input_tokens, system_tokens, extra_tokens, ctx_tokens, ctx_len, ctx_len_all, \
               sum_tokens, max_current, threshold)
        """
        model = self.window.core.config.get('model')
        model_id = self.window.core.models.get_id(model)
        mode = self.window.core.config.get('mode')
        user_name = self.window.core.config.get('user_name')
        ai_name = self.window.core.config.get('ai_name')

        system_prompt = ""
        system_tokens = 0
        input_tokens = 0
        max_total_tokens = self.window.core.config.get('max_total_tokens')
        extra_tokens = self.get_extra(model)

        if mode in CHAT_MODES:
            # system prompt (without extra tokens)
            system_prompt = str(self.window.core.config.get('prompt')).strip()
            system_prompt = self.window.core.prompt.build_final_system_prompt(system_prompt)  # add addons

            if system_prompt is not None and system_prompt != "":
                system_tokens = self.from_prompt(system_prompt, "", model_id)
                system_tokens += self.from_text("system", model_id)

            # input prompt
            if input_prompt is not None and input_prompt != "":
                input_tokens = self.from_prompt(input_prompt, "", model_id)
                input_tokens += self.from_text("user", model_id)
        elif mode == "completion":
            # system prompt (without extra tokens)
            system_prompt = str(self.window.core.config.get('prompt')).strip()
            system_prompt = self.window.core.prompt.build_final_system_prompt(system_prompt)  # add addons
            system_tokens = self.from_text(system_prompt, model_id)

            # input prompt
            if input_prompt is not None and input_prompt != "":
                message = ""
                if user_name is not None \
                        and ai_name is not None \
                        and user_name != "" \
                        and ai_name != "":
                    message += "\n" + user_name + ": " + str(input_prompt)
                    message += "\n" + ai_name + ":"
                else:
                    message += "\n" + str(input_prompt)
                input_tokens = self.from_text(message, model_id)
                extra_tokens = 0  # no extra tokens in completion mode

        # tmp system prompt
        self.window.core.ctx.current_sys_prompt = system_prompt

        # used tokens
        used_tokens = system_tokens + input_tokens

        # check model max allowed ctx tokens
        max_current = max_total_tokens
        model_ctx = self.window.core.models.get_num_ctx(model_id)
        if max_current > model_ctx:
            max_current = model_ctx

        # context threshold (reserved for output)
        threshold = self.window.core.config.get('context_threshold')
        max_to_check = max_current - threshold

        # context tokens
        ctx_len_all = len(self.window.core.ctx.items)
        ctx_len, ctx_tokens = self.window.core.ctx.count_prompt_items(model_id, mode, used_tokens, max_to_check)

        # empty ctx tokens if context is not used
        if not self.window.core.config.get('use_context'):
            ctx_tokens = 0
            ctx_len = 0

        # sum of input tokens
        sum_tokens = system_tokens + input_tokens + ctx_tokens + extra_tokens

        return input_tokens, system_tokens, extra_tokens, ctx_tokens, ctx_len, ctx_len_all, \
               sum_tokens, max_current, threshold

    def from_user(self, system_prompt: str, input_prompt: str) -> int:
        """
        Count per-user used tokens

        :param system_prompt: system prompt
        :param input_prompt: input prompt
        :return: used tokens
        """
        model = self.window.core.config.get('model')
        model_id = self.window.core.models.get_id(model)
        mode = self.window.core.config.get('mode')
        tokens = 0
        if mode == "chat" or mode == "vision":
            tokens += self.from_prompt(system_prompt, "", model_id)  # system prompt
            tokens += self.from_text("system", model_id)
            tokens += self.from_prompt(input_prompt, "", model_id)  # input prompt
            tokens += self.from_text("user", model_id)
        else:
            # rest of modes
            tokens += self.from_text(system_prompt, model_id)  # system prompt
            tokens += self.from_text(input_prompt, model_id)  # input prompt
        tokens += self.window.core.config.get('context_threshold')  # context threshold (reserved for output)
        tokens += self.get_extra(model_id)  # extra tokens (required for output)
        return tokens

    @staticmethod
    def get_config(model: str) -> (str, int, int):
        """
        Return tokens config values

        :param model: model ID
        :return: model, per_message, per_name
        """
        per_message = 4  # message follows <|start|>{role/name}\n{content}<|end|>\n
        per_name = -1

        if model is not None:
            if model in {
                "gpt-3.5-turbo-0613",
                "gpt-3.5-turbo-16k-0613",
                "gpt-4-0314",
                "gpt-4-32k-0314",
                "gpt-4-0613",
                "gpt-4-32k-0613",
            }:
                per_message = 3
                per_name = 1
            elif model == "gpt-3.5-turbo-0301":
                per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
                per_name = -1  # if there's a name, the role is omitted
            elif "gpt-3.5-turbo" in model:
                return Tokens.get_config(model="gpt-3.5-turbo-0613")
            elif "gpt-4" in model:
                return Tokens.get_config(model="gpt-4-0613")
            elif model.startswith("text-davinci"):
                per_message = 1
                per_name = 0

        return model, per_message, per_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.06 02:00:00                  #
# ================================================== #

import os
import shutil
import time

from sqlalchemy import create_engine, text

from pygpt_net.migrations import Migrations
from .viewer import Viewer


class Database:
    def __init__(self, window=None):
        """Database provider core"""
        self.window = window
        self.db_path = None
        self.db_name = 'db.sqlite'
        self.engine = None
        self.initialized = False
        self.echo = True
        columns = {}
        columns["calendar_note"] = [
            'id',
            'idx',
            'status',
            'year',
            'month',
            'day',
            'created_ts',
            'updated_ts',
            'title',
            'content',
            'is_important',
            'is_deleted',
            'uuid',
        ]
        columns["config"] = [
            'config_key',
            'config_value',
            'created_ts',
            'updated_ts',
        ]
        columns["ctx_item"] = [
            'id',
            'meta_id',
            'input',
            'output',
            'input_name',
            'output_name',
            'input_ts',
            'output_ts',
            'mode',
            'model',
            'thread_id',
            'msg_id',
            'run_id',
            'cmds_json',
            'results_json',
            'urls_json',
            'images_json',
            'files_json',
            'attachments_json',
            'extra',
            'input_tokens',
            'output_tokens',
            'total_tokens',
            'is_internal',
            'docs_json',
            'external_id',
        ]
        columns["ctx_meta"] = [
            'id',
            'name',
            'updated_ts',
            'mode',
            'model',
            'created_ts',
            'indexed_ts',
            'last_mode',
            'last_model',
            'thread_id',
            'assistant_id',
            'preset_id',
            'run_id',
            'status',
            'extra',
            'is_initialized',
            'is_deleted',
            'is_important',
            'is_archived',
            'label',
            'indexes_json',
            'external_id',
            'uuid',
        ]
        columns["idx_ctx"] = [
            'id',
            'meta_id',
            'updated_ts',
            'item_id',
            'doc_id',
            'created_ts',
            'store',
            'idx',
            'uuid',
        ]
        columns["idx_external"] = [
            'id',
            'content',
            'type',
            'doc_id',
            'created_ts',
            'updated_ts',
            'store',
            'idx',
            'uuid',
        ]
        columns["idx_file"] = [
            'id',
            'name',
            'path',
            'doc_id',
            'created_ts',
            'updated_ts',
            'store',
            'idx',
            'uuid',
        ]
        columns["notepad"] = [
            'id',
            'idx',
            'title',
            'content',
            'created_ts',
            'updated_ts',
            'is_deleted',
            'is_initialized',
            'uuid',
        ]

        self.tables = {
            'calendar_note': {
                'columns': columns["calendar_note"],
                'sort_by': columns["calendar_note"],
                'search_fields': ['id', 'title', 'content'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'config': {
                'columns': columns["config"],
                'sort_by': columns["config"],
                'search_fields': ['config_key', 'config_value'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'config_key',
                'default_order': 'ASC',
                'primary_key': 'config_key',
            },
            'ctx_item': {
                'columns': columns["ctx_item"],
                'sort_by': columns["ctx_item"],
                'search_fields': ['id', 'input', 'output', 'input_name', 'output_name', 'meta_id'],
                'timestamp_columns': ['input_ts', 'output_ts'],
                'json_columns': ['cmds_json', 'results_json', 'urls_json', 'images_json', 'files_json', 'attachments_json', 'docs_json'],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'ctx_meta': {
                'columns': columns["ctx_meta"],
                'sort_by': columns["ctx_meta"],
                'search_fields': ['id', 'name'],
                'timestamp_columns': ['created_ts', 'updated_ts', 'indexed_ts'],
                'json_columns': ['indexes_json'],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'idx_ctx': {
                'columns': columns["idx_ctx"],
                'sort_by': columns["idx_ctx"],
                'search_fields': ['id', 'doc_id', 'meta_id', 'item_id', 'store', 'idx'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'idx_file': {
                'columns': columns["idx_file"],
                'sort_by': columns["idx_file"],
                'search_fields': ['id', 'doc_id', 'name', 'path', 'store', 'idx'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'idx_external': {
                'columns': columns["idx_external"],
                'sort_by': columns["idx_external"],
                'search_fields': ['id', 'doc_id', 'content', 'type', 'store', 'idx'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
            'notepad': {
                'columns': columns["notepad"],
                'sort_by': columns["notepad"],
                'search_fields': ['id', 'title', 'content'],
                'timestamp_columns': ['created_ts', 'updated_ts'],
                'json_columns': [],
                'default_sort': 'id',
                'default_order': 'DESC',
                'primary_key': 'id',
            },
        }
        self.migrations = Migrations()
        self.viewer = Viewer(self)

    def init(self):
        """Initialize database"""
        if not self.initialized:
            self.db_path = os.path.join(self.window.core.config.path, self.db_name)
            self.prepare()

    def get_tables(self) -> dict:
        """
        Get tables configuration

        :return: tables
        """
        return self.tables

    def prepare(self):
        """Prepare database"""
        self.engine = create_engine(
            'sqlite:///{}'.format(self.db_path),
            echo=self.echo,
            future=True
        )
        if not self.is_installed():
            self.install()
        self.initialized = True

    def install(self):
        """Install database schema"""
        with self.engine.begin() as conn:
            conn.execute(text("""
            CREATE TABLE IF NOT EXISTS config (
                config_key TEXT PRIMARY KEY,
                config_value TEXT,
                created_ts INTEGER,
                updated_ts INTEGER
            );"""))
            print("[DB] Installed database: {}".format(self.db_path))

    def get_version(self) -> int:
        """
        Get database migration version

        :return: version string
        """
        return int(self.get_param("db_version") or 0)

    def get_db(self):
        """
        Get database engine

        :return: database engine
        """
        return self.engine

    def is_installed(self) -> bool:
        """
        Check if database is installed

        :return: True if installed
        """
        if not os.path.exists(self.db_path):
            return False

        with self.engine.connect() as conn:
            result = conn.execute(text("""
            SELECT COUNT(*) FROM sqlite_master WHERE type='table' AND name='config';
            """)).fetchone()
            return result[0] == 1

    def apply_migration(self, migration, conn, db_version: int):
        """
        Apply DB migration

        :param migration: migration object
        :param conn: database connection
        :param db_version: database version
        """
        migration.window = self.window
        migration_version = int(migration.__class__.__name__.replace('Version', ''))
        if migration_version > db_version:
            migration.up(conn)
            self.set_param_exec("db_version", migration_version, conn)
            print("[DB] Executed DB migration: {}".format(migration.__class__.__name__).replace('Version', ''))

    def has_migrations_to_apply(self, migrations) -> bool:
        """
        Check if there are any migrations to apply
        
        :param migrations: list of migrations
        """
        has_migrations = False
        for migration in migrations:
            migration.window = self.window
            migration_version = int(migration.__class__.__name__.replace('Version', ''))
            if migration_version > self.get_version():
                has_migrations = True
                break
        return has_migrations

    def make_backup(self) -> str or None:
        """
        Make backup of database before migration

        :return: backup path
        """
        try:
            backup_path = os.path.join(self.window.core.config.path, 'db.sqlite.backup')
            if os.path.exists(backup_path):
                os.remove(backup_path)
            shutil.copyfile(self.db_path, backup_path)
            return backup_path
        except Exception as e:
            print("[DB] Error while making backup of database: {}".format(e))
        return None

    def migrate(self):
        """Apply all DB migrations"""
        db_version = self.get_version()
        migrations = Migrations().get_versions()
        sorted_migrations = sorted(migrations, key=lambda m: m.__class__.__name__)
        self.init()
        if self.has_migrations_to_apply(sorted_migrations):
            print("[DB] Migrating database...")
            self.make_backup()  # make backup of current database
            with self.engine.begin() as conn:
                for migration in sorted_migrations:
                    self.apply_migration(migration, conn, db_version)

    def get_param(self, key: str) -> str or None:
        """
        Get parameter from database

        :param key: parameter key
        :return: parameter value
        """
        self.init()

        with self.engine.connect() as conn:
            sel_stmt = text("SELECT config_value FROM config WHERE config_key = :key").bindparams(key=key)
            result = conn.execute(sel_stmt).fetchone()
            return result[0] if result else None

    def set_param(self, key: str, value: any):
        """
        Insert or update parameter in database

        :param key: parameter key
        :param value: parameter value
        """
        self.init()
        with self.engine.begin() as conn:
            self.set_param_exec(key, value, conn)

    def set_param_exec(self, key: str, value: any, conn):
        """
        Insert or update parameter in database

        :param key: parameter key
        :param value: parameter value
        :param conn: database connection
        """
        ts = int(time.time())
        stmt = text("SELECT 1 FROM config WHERE config_key = :key").bindparams(key=key)
        result = conn.execute(stmt).fetchone()

        if result:
            stmt = text("""
                UPDATE config
                SET config_value = :value, updated_ts = :updated_ts
                WHERE config_key = :key
            """).bindparams(key=key, value=value, updated_ts=ts)
        else:
            stmt = text("""
                INSERT INTO config (config_key, config_value, created_ts, updated_ts) 
                VALUES (:key, :value, :created_ts, :updated_ts)
            """).bindparams(key=key, value=value, created_ts=ts, updated_ts=ts)

        conn.execute(stmt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.28 01:00:00                  #
# ================================================== #

from datetime import datetime
from logging import ERROR, WARNING, INFO, DEBUG

from PySide6.QtCore import Slot
from PySide6.QtGui import QTextCursor


class Debug:
    def __init__(self, window=None):
        """
        Debug controller

        :param window: Window instance
        """
        self.window = window
        self.is_logger = False  # logger window opened
        self.is_app_log = False  # app log window opened
        self.allow_level_change = False  # allow changing log level

    def update(self):
        """Update debug"""
        self.update_menu()

    def update_menu(self):
        """Update debug menu"""
        for id in self.window.controller.dialogs.debug.get_ids():
            if self.window.controller.dialogs.debug.is_active(id):
                self.window.ui.menu['debug.' + id].setChecked(True)
            else:
                self.window.ui.menu['debug.' + id].setChecked(False)

        if self.is_logger:
            self.window.ui.menu['debug.logger'].setChecked(True)
        else:
            self.window.ui.menu['debug.logger'].setChecked(False)

        if self.is_app_log:
            self.window.ui.menu['debug.app.log'].setChecked(True)
        else:
            self.window.ui.menu['debug.app.log'].setChecked(False)

    def set_log_level(self, level: str = 'error'):
        """
        Switch logging level in runtime

        :param level: log level (debug, info, warning, error), default: error
        """
        if not self.allow_level_change:
            return

        print("[LOGGER] Changing log level to: " + level)

        if level == 'debug':
            self.window.core.debug.switch_log_level(DEBUG)
            print("** DEBUG level enabled")
        elif level == 'info':
            self.window.core.debug.switch_log_level(INFO)
            print("** INFO level enabled")
        elif level == 'warning':
            self.window.core.debug.switch_log_level(WARNING)
            print("** WARNING level enabled")
        else:
            self.window.core.debug.switch_log_level(ERROR)
            print("** ERROR level enabled")

        self.window.ui.dialogs.app_log.update_log_level()

    def on_update(self, all: bool = False):
        """
        Update debug windows (only if active)

        :param all: update all debug windows
        """
        # not_realtime = ['context']
        not_realtime = []
        for id in self.window.controller.dialogs.debug.get_ids():
            if self.window.controller.dialogs.debug.is_active(id):
                if all or id not in not_realtime:
                    self.window.controller.dialogs.debug.update_worker(id)

    def setup(self):
        """Setup debug"""
        current = self.window.core.debug.get_log_level()
        if current == ERROR:
            self.allow_level_change = True
        else:
            return

        # switch log level if set in config
        if self.window.core.config.has('log.level'):
            level = self.window.core.config.get('log.level')
            if level != "error":
                print("[LOGGER] Started with log level: " + self.window.core.debug.get_log_level_name())
                print("[LOGGER] Switching to: " + level)
                self.set_log_level(level)

    @Slot(object)
    def handle_log(self, data: any):
        """
        Handle log message

        :param data: message to log
        """
        self.log(data)

    def log(self, data: any, window: bool = True):
        """
        Log message to console or logger window

        :param data: text to log
        :param window: True if log to window, False if log to console
        """
        if not window:
            print(str(data))
            return

        if not self.is_logger or data is None or str(data).strip() == "":
            return

        data = datetime.now().strftime('%H:%M:%S.%f') + ': ' + str(data)
        cur = self.window.logger.textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(data) + "\n"
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertText("\n")
        self.window.logger.setTextCursor(cur)  # Update visible cursor

    def logger_enabled(self) -> bool:
        """
        Check if debug window is enabled

        :return: True if enabled, False otherwise
        """
        return self.is_logger

    def open_logger(self):
        """Open logger dialog"""
        self.window.ui.dialogs.open('logger', width=800, height=600)
        self.is_logger = True
        self.update()

    def close_logger(self):
        """Close logger dialog"""
        self.window.ui.dialogs.close('logger')
        self.is_logger = False
        self.update()

    def toggle_logger(self):
        """Toggle logger dialog"""
        if self.is_logger:
            self.close_logger()
        else:
            self.open_logger()

    def clear_logger(self):
        """Clear logger dialog"""
        self.window.logger.clear()

    def toggle(self, id: str):
        """
        Toggle debug window

        :param id: debug window to toggle
        """
        if self.window.controller.dialogs.debug.is_active(id):
            self.window.controller.dialogs.debug.hide(id)
        else:
            self.window.controller.dialogs.debug.show(id)
            self.on_update(True)

        self.log('debug.' + id + ' toggled')

        # update menu
        self.update()

    def toggle_app_log(self):
        """
        Toggle app log window
        """
        id = 'app.log'
        if self.is_app_log:
            self.window.ui.dialogs.close(id)
            self.is_app_log = False
        else:
            self.window.ui.dialogs.open(id, width=800, height=600)
            self.window.ui.dialogs.app_log.reload()
            self.is_app_log = True

        self.log('debug.' + id + ' toggled')

        # update menu
        self.update()
